/* Particles: jets, explosions, and space dust.
 * Copyright (C) 2004-2007 Jason Woofenden and Joshua Grams.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "config.h"
#include "globals.h"
#include "common.h"
#include "dots.h"
#include "float.h"
#include "mt.h"

struct dust_mote {
	float x,y,z;
	Uint16 color;
};

static struct dust_mote motes[N_DUST_MOTES];

struct dot {
	int active;
	float x, y;
	float dx, dy;
	float mass;   // in DOT_MASS_UNITs
	float decay;  // rate at which to reduce mass.
	int heat;     // heat multiplier (color).
};

static struct dot edot[MAXENGINEDOTS];
static struct dot bdot[MAXBANGDOTS];

static uint16_t heatcolor[W*3];


static void
init_dust(void)
{
	int i, b;
	for(i=0; i<N_DUST_MOTES; i++) {
		motes[i].x = frnd()*(XSIZE-5);
		motes[i].y = frnd()*(YSIZE-5);
		motes[i].z = MAX_DUST_DEPTH*sqrt(frnd());
		b = (MAX_DUST_DEPTH - motes[i].z) * 255.0 / MAX_DUST_DEPTH;
		motes[i].color = SDL_MapRGB(surf_screen->format, b, b, b);
	}
}

void
init_dots(void)
{
	int i;

	init_dust();

	// Set the heat colors from the range 0 (cold) to 300 (blue-white)
	for(i = 0; i<W*3; i++) {
		heatcolor[i] = SDL_MapRGB(
			surf_screen->format,
			(i<W)?(i*M/W):(M),(i<W)?0:(i<2*W)?((i-W)*M/W):M,(i<2*W)?0:((i-W)*M/W) // Got that?
		);
	}
}


void
new_engine_dots(struct ship *ship)
{
	static int dot = 0;
	int dir, i;
	int n = t_frame * ENGINE_DOTS_PER_TIC;
	float a, r;  // angle, random length
	float dx, dy;
	float hx, hy; // half ship width/height.
	static const int s[4] = { 2, 1, 0, 1 };
	float time;
	float accelh, accelv, past_ship_dx, past_ship_dy;

	hx = ship->image->w / 2;
	hy = ship->image->h / 2;

	for(dir=0; dir<4; dir++) {
		if(!(ship->jets & 1<<dir)) continue;

		for(i = 0; i<n; i++) {
			if(!edot[dot].active) {
				a = frnd()*M_PI + (dir-1)*M_PI_2;
				r = sin(frnd()*M_PI);
				dx = r * cos(a);
				dy = r * -sin(a);  // screen y is "backwards".

				edot[dot].active = 1;
				edot[dot].decay = 3;
				edot[dot].heat = 6;

				// dot was created at a random time during the time span
				time = frnd() * t_frame; // this is how long ago

				// calculate how fast the ship was going when this engine dot was
				// created (as if it had a smooth acceleration). This is used in
				// determining the velocity of the dots, but not their starting
				// location.
				accelh = ((ship->jets >> 2) & 1) - (ship->jets & 1);
				accelh *= THRUSTER_STRENGTH * time;
				past_ship_dx = ship->dx - accelh;
				accelv = ((ship->jets >> 1) & 1) - ((ship->jets >> 3) & 1);
				accelv *= THRUSTER_STRENGTH * time;
				past_ship_dy = ship->dy - accelv;

				// the starting position (not speed) of the dot is calculated as
				// though the ship were traveling at a constant speed for this
				// t_frame.
				edot[dot].x = (ship->x - (ship->dx - screendx) * time) + s[dir]*hx;
				edot[dot].y = (ship->y - (ship->dy - screendy) * time) + s[(dir+1)&3]*hy;
				if(dir&1) {
					edot[dot].dx = past_ship_dx + 2*dx;
					edot[dot].dy = past_ship_dy + 20*dy;
					edot[dot].mass = 60 * fabs(dy);
				} else {
					edot[dot].dx = past_ship_dx + 20*dx;
					edot[dot].dy = past_ship_dy + 2*dy;
					edot[dot].mass = 60 * fabs(dx);
				}

				// move the dot as though it were created in the past
				edot[dot].x += (edot[dot].dx - screendx) * time;
				edot[dot].y += (edot[dot].dy - screendy) * time;

				dot = (dot+1 % MAXENGINEDOTS);
			}
		}
	}
}


void
explode(Sprite *s)
{
	static int dot = 0;  // dot index
	int times, x, y;   // loop counters
	double theta, r;   // random numbers
	uint16_t *pixel, c;  // the image
	uint32_t colorkey;
	int row_inc;
	SDL_Surface *i = s->image;

	pixel = i->pixels;
	row_inc = i->pitch/sizeof(uint16_t) - i->w;
	colorkey = i->format->colorkey;

	if(SDL_MUSTLOCK(i)) { SDL_LockSurface(i); }

	for(times=0; times<20; times++) {
		pixel = i->pixels;
		for(y=0; y<i->h; y++) {
			for(x = 0; x<i->w; x++) {
				c = *pixel++;
				if(c && c != colorkey) {
					theta = frnd()*M_PI*2;
					r = frnd(); r = 1 - r*r;

					bdot[dot].dx = 45*r*cos(theta) + s->dx;
					bdot[dot].dy = 45*r*sin(theta) + s->dy;
					bdot[dot].x = x + s->x;
					bdot[dot].y = y + s->y;
					bdot[dot].mass = frnd() * 99;
					bdot[dot].decay = frnd()*1.5 + 0.5;
					bdot[dot].heat = 3;
					bdot[dot].active = 1;

					dot = (dot+1) % MAXBANGDOTS;
				}
				pixel += row_inc;
			}
		}
	}

	if(SDL_MUSTLOCK(i)) { SDL_UnlockSurface(i); }
}


static void
move_dust(void)
{
	int i;
	float xscroll = screendx * t_frame;
	float yscroll = screendy * t_frame;
	
	for(i=0; i<N_DUST_MOTES; i++) {
		motes[i].x -= xscroll / (1.3 + motes[i].z);
		motes[i].x = fwrap(motes[i].x, XSIZE);

		motes[i].y -= yscroll / (1.3 + motes[i].z);
		motes[i].y = fwrap(motes[i].y, YSIZE);
	}
}

static void
move_dot(struct dot *d)
{
	Sprite *hit;
	float mass;

	if(d->active) {
		d->x += (d->dx - screendx) * t_frame;
		d->y += (d->dy - screendy) * t_frame;
		d->mass -= t_frame * d->decay;
		if(d->mass < 0 || fclip(d->x, XSIZE) || fclip(d->y, YSIZE))
			d->active = 0; 
		else {
			hit = pixel_collides(d->x, d->y);
			if(hit && hit->type != SHIP) {
				d->active = 0;
				mass = sprite_mass(hit);
				hit->dx += DOT_MASS_UNIT * d->mass * d->dx / mass;
				hit->dy += DOT_MASS_UNIT * d->mass * d->dy / mass;
			}
		}
	}
}

void
move_dots(void)
{
	int i;

	move_dust();
	for(i=0; i<MAXBANGDOTS; i++) move_dot(&bdot[i]);
	for(i=0; i<MAXENGINEDOTS; i++) move_dot(&edot[i]);
}


static void
draw_dust(void)
{
	int i;
	uint16_t *pixels = surf_screen->pixels;
	for(i=0; i<N_DUST_MOTES; i++) {
		pixels[surf_screen->pitch/2*(int)motes[i].y + (int)motes[i].x] = motes[i].color;
	}
}

static void
draw_dot(struct dot *d)
{
	uint16_t *pixels, *pixel;
	int row_inc;

	if(d->active) {
		pixels = (uint16_t *) surf_screen->pixels;
		row_inc = surf_screen->pitch / sizeof(uint16_t);
		pixel = pixels + (int)d->y * row_inc + (int)d->x;
		*pixel = heatcolor[min(3*W-1, (int)(d->mass * d->heat))];
	}
}

void
draw_dots(void)
{
	int i;

	if(SDL_MUSTLOCK(surf_screen)) { SDL_LockSurface(surf_screen); }
	draw_dust();
	for(i=0; i<MAXBANGDOTS; i++) draw_dot(&bdot[i]);
	for(i=0; i<MAXENGINEDOTS; i++) draw_dot(&edot[i]);
	if(SDL_MUSTLOCK(surf_screen)) { SDL_UnlockSurface(surf_screen); }
}
