/* Variations on RockDodger
 * Space Rocks copyright (C) 2001 Paul Holt <pad@pcholt.com>
 *
 * Project fork 2004, Jason Woofenden and Joshua Grams.
 * (a whole bunch of modifications and project rename)

 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <math.h>
#include <SDL.h>
#include <SDL_image.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "globals.h"

#include "font.h"

#include "args.h"
#include "common.h"
#include "dots.h"
#include "file.h"
#include "float.h"
#include "mt.h"
#include "rocks.h"
#include "score.h"
#include "sprite.h"
#include "sound.h"

// ************************************* VARS
// SDL_Surface global variables
SDL_Surface 
	*surf_screen,	// Screen
	*surf_b_variations, // "variations" banner
	*surf_b_on, // "on" banner
	*surf_b_rockdodger, // "rockdodger" banner
	*surf_b_game,	// Title element "game"
	*surf_b_over,	// Title element "over"
	*surf_life,	// Indicator of number of ships remaining
	*surf_rock[NROCKS],	// THE ROCKS
	*surf_font_big;	// The big font
	

font *g_font;

// Other global variables
char topline[1024];
char *initerror = "";

struct ship ship = { SHIP, 0, NULL, XSIZE/2, YSIZE/2, BARRIER_SPEED, 0.0 };
	  
float screendx = BARRIER_SPEED, screendy = 0.0;
float dist_ahead;

// all movement is based on t_frame.
unsigned long frames, start, end;
float t_frame;  // length of this frame (in ticks = 1/20th second)  adjusted for gamespeed
int ms_frame;   // length of this frame (milliseconds)
int ms_end;     // end of this frame (milliseconds)

float gamespeed = 1.00;

int score;

float fadetimer = 0;

int pausedown = 0, paused = 0;

enum states {
	TITLE_PAGE,
	GAMEPLAY,
	DEAD_PAUSE,
	GAME_OVER,
	HIGH_SCORE_ENTRY,
	HIGH_SCORE_DISPLAY
};
enum states state = TITLE_PAGE;
float state_timeout = 600.0;

#define NSEQUENCE 3
char *msgs[2][3] = {
	{
		"Press SPACE for normal game",
		"Press '1' for easy game",
		"http://jasonwoof.org/vor"
	},
	{
		"Press SPACE for easy game",
		"Press '2' for normal game",
		"http://jasonwoof.org/vor"
	}
};

char *data_dir;
extern char *optarg;
extern int optind, opterr, optopt;

#define TO_TICKS(seconds) ((seconds)*20*gamespeed)

// ************************************* FUNCS

SDL_Surface *
load_image(char *filename)
{
	SDL_Surface *tmp, *img = NULL;
	char *s = add_data_path(filename);
	if(s) {
		tmp = IMG_Load(s);
		free(s);
		if(tmp) {
			img = SDL_DisplayFormat(tmp);
			SDL_FreeSurface(tmp);
		}
	}
	return img;
}

void
load_ship(void)
{
	load_sprite(SPRITE(&ship), "ship.png");
}

void font_cleanup() {
	font_free(g_font);
}

int
init(void) {

	char *s;
	Uint32 flag;

	// Where are our data files?
	if(!find_files()) exit(1);
	read_high_score_table();

	if(opt_sound) {
		// Initialize SDL with audio and video
		if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) != 0) {
			opt_sound = 0;
			fputs("Can't open sound, starting without it\n", stderr);
			atexit(SDL_Quit);
		} else {
			atexit(SDL_Quit);
			atexit(SDL_CloseAudio);
			opt_sound = init_sound();
		}
	} else {
		// Initialize with video only
		CONDERROR(SDL_Init(SDL_INIT_VIDEO) != 0);
		atexit(SDL_Quit);
	}

	play_tune(TUNE_TITLE_PAGE);

	// Attempt to get the required video size
	flag = SDL_DOUBLEBUF | SDL_HWSURFACE;
	if(opt_fullscreen) flag |= SDL_FULLSCREEN;
	surf_screen = SDL_SetVideoMode(XSIZE,YSIZE,16,flag);

	// Set the title bar text
	SDL_WM_SetCaption("Variations on Rockdodger", "VoR");

	NULLERROR(surf_screen);

	// Load the banners
	NULLERROR(surf_b_variations = load_image("b_variations.png"));
	NULLERROR(surf_b_on = load_image("b_on.png"));
	NULLERROR(surf_b_rockdodger = load_image("b_rockdodger.png"));

	NULLERROR(surf_b_game = load_image("b_game.png"));
	NULLERROR(surf_b_over = load_image("b_over.png"));

	// Load the life indicator (small ship) graphic.
	NULLERROR(surf_life = load_image("life.png"));

	// Load the font image
	s = add_data_path("font.png");
	if(s) {
		g_font = font_load(s);
		atexit(&font_cleanup);
	} else {
		fprintf(stderr, "could create path to font\n");
		exit(1);
	}

	init_dots();

	init_sprites();
	add_sprite(SPRITE(&ship));

	// Remove the mouse cursor
#ifdef SDL_DISABLE
	SDL_ShowCursor(SDL_DISABLE);
#endif

	return 0;
}

void
show_lives(void)
{
	int i;
	SDL_Rect dest;

	for(i=0; i<ship.lives-1; i++) {
		dest.x = (i + 1)*(surf_life->w + 10);
		dest.y = 20;
		SDL_BlitSurface(surf_life, NULL, surf_screen, &dest);
	}
}

void
draw_game_over(void)
{
	int x;
	char *text0, *text1;
	SDL_Rect dest;

	fadetimer += t_frame;

	dest.x = (XSIZE-surf_b_game->w)/2;
	dest.y = (YSIZE-surf_b_game->h)/2-40;
	SDL_BlitSurface(surf_b_game,NULL,surf_screen,&dest);

	dest.x = (XSIZE-surf_b_over->w)/2;
	dest.y = (YSIZE-surf_b_over->h)/2 + 40;
	SDL_BlitSurface(surf_b_over,NULL,surf_screen,&dest);

	if(new_high_score(score)) {
		text0 = "New High Score!";
		text1 = "Press SPACE to continue";
	} else {
		text0 = msgs[g_easy][0];
		text1 = msgs[g_easy][1];
	}

	x = (XSIZE-font_width(text0))/2 + cos(fadetimer/9)*10;
	font_write(x,YSIZE-100 + cos(fadetimer/6)*5,text0);

	x = (XSIZE-font_width(text1))/2 + sin(fadetimer/9)*10;
	font_write(x,YSIZE-50 + sin(fadetimer/4)*5,text1);
}

void
draw_title_page(void)
{
	int x;
	char *text;
	SDL_Rect dest;

	fadetimer += t_frame/2.0;

	dest.x = (XSIZE-surf_b_variations->w)/2 + cos(fadetimer/6.5)*10;
	dest.y = (YSIZE/2-surf_b_variations->h)/2 + sin(fadetimer/5.0)*10;
	SDL_BlitSurface(surf_b_variations,NULL,surf_screen,&dest);

	dest.x = (XSIZE-surf_b_on->w)/2 + cos((fadetimer + 1.0)/6.5)*10;
	dest.y = (YSIZE/2-surf_b_on->h)/2 + surf_b_variations->h + 20 + sin((fadetimer + 1.0)/5.0)*10;
	SDL_BlitSurface(surf_b_on,NULL,surf_screen,&dest);

	dest.x = (XSIZE-surf_b_rockdodger->w)/2 + cos((fadetimer + 2.0)/6.5)*10;
	dest.y = (YSIZE/2-surf_b_rockdodger->h)/2 + surf_b_variations->h + surf_b_on->h + 40 + sin((fadetimer + 2.0)/5)*10;
	SDL_BlitSurface(surf_b_rockdodger,NULL,surf_screen,&dest);

	text = msgs[g_easy][(int)(fadetimer/35)%NSEQUENCE];
	x = (XSIZE-font_width(text))/2 + cos(fadetimer/4.5)*10;
	font_write(x,YSIZE-100 + cos(fadetimer/3)*5,text);

	text = "Version " VERSION;
	x = (XSIZE-font_width(text))/2 + sin(fadetimer/4.5)*10;
	font_write(x,YSIZE-50 + sin(fadetimer/2)*5,text);
}

void
draw(void)
{
	SDL_FillRect(surf_screen,NULL,0);  // black background
	draw_dots();            // background dots
	draw_sprite(SPRITE(&ship));
	draw_rocks();

	show_lives();
	show_score();

	switch (state) {
		case GAME_OVER: draw_game_over(); break;

		case TITLE_PAGE: draw_title_page(); break;

		case HIGH_SCORE_ENTRY: play_tune(TUNE_HIGH_SCORE_ENTRY);
			// and fall through to
		case HIGH_SCORE_DISPLAY:
			// Display de list o high scores mon.
			display_scores(150,50);
			break;
		case GAMEPLAY:
		case DEAD_PAUSE:
			; // no action necessary
	}

	// Update the surface
	SDL_Flip(surf_screen);
}

static inline void
kill_ship(struct ship *ship)
{
	play_sound(SOUND_BANG);
	explode(SPRITE(ship));
	if(--ship->lives) {
		state = DEAD_PAUSE;
		state_timeout = DEAD_PAUSE_LENGTH;
		// want ship to be invisible, but keep drifting at sqrt(speed)
		// to leave it in the middle of the space from the explosion.
		ship->flags = MOVE;
		ship->dx = (ship->dx < 0) ? -sqrt(-ship->dx) : sqrt(ship->dx);
		ship->dy = (ship->dy < 0) ? -sqrt(-ship->dy) : sqrt(ship->dy);
		if(ship->dx < BARRIER_SPEED) ship->dx = BARRIER_SPEED;
	} else {
		state = GAME_OVER;
		state_timeout = 200.0;
		fadetimer = 0.0;
		ship->flags = 0;
		// scrolling is based on the ship speed, so we need to reset it.
		ship->dx = BARRIER_SPEED; ship->dy = 0;
	}
}

void
do_collision(Sprite *a, Sprite *b)
{
	if(a->type == SHIP) kill_ship((struct ship *)a);
	else if(b->type == SHIP) kill_ship((struct ship *)b);
	else bounce(a, b);
}

void
init_score_entry(void)
{
	SDL_Event e;
	state = HIGH_SCORE_ENTRY;
	state_timeout = 5.0e6;
	SDL_EnableUNICODE(1);
	while(SDL_PollEvent(&e))
		;
	insert_score(score);
}

// Count down the state timer, and change state when it gets to zero or less;
void
update_state(void)
{
	state_timeout -= t_frame*3;
	if(state_timeout > 0) return;

	switch(state) {
		case GAMEPLAY: break;  // no action necessary
		case DEAD_PAUSE:
			// Restore the ship and continue playing
			ship.flags = DRAW|MOVE|COLLIDE;
			state = GAMEPLAY;
			play_tune(TUNE_GAMEPLAY);
			break;
		case GAME_OVER:
			if(new_high_score(score)) init_score_entry();
			else {
				state = HIGH_SCORE_DISPLAY;
				state_timeout = 400;
			}
			break;
		case HIGH_SCORE_DISPLAY:
			state = TITLE_PAGE;
			state_timeout = 600.0;
			fadetimer = 0.0;
			break;
		case HIGH_SCORE_ENTRY:
			break;
		case TITLE_PAGE:
			state = HIGH_SCORE_DISPLAY;
			state_timeout = 200.0;
			break;
	}
}

void
gameloop() {
	SDL_Event e;
	Uint8 *keystate;
	float tmp;

	for(;;) {
		ms_frame = SDL_GetTicks() - ms_end;
		ms_end += ms_frame;
		if(ms_frame > 1000) {
			ms_frame = 1000;
		}
		t_frame = gamespeed * ms_frame / 50;
		frames++;

		while(SDL_PollEvent(&e)) {
			switch(e.type) {
				case SDL_QUIT: return;
				case SDL_KEYDOWN:
					if(state == HIGH_SCORE_ENTRY) {
						if(!process_score_input(&e.key.keysym)) {
							// Write the high score table to the file
							write_high_score_table();
							// continue to display the scores briefly
							state = HIGH_SCORE_DISPLAY;
							state_timeout = 200;
							play_tune(TUNE_TITLE_PAGE);
						}
					} else if(e.key.keysym.sym == SDLK_q) {
						return;
					}

					if(e.key.keysym.sym == SDLK_ESCAPE) {
						return;
					}
					break;
			}
		}
		keystate = SDL_GetKeyState(NULL);

		if(state == GAMEPLAY) {
			if(!paused) {
				score += ms_frame;
				
				if(keystate[SDLK_LEFT]  || keystate[SDLK_h]) { ship.dx -= THRUSTER_STRENGTH*t_frame; ship.jets |= 1<<0;}
				if(keystate[SDLK_DOWN]  || keystate[SDLK_t]) { ship.dy += THRUSTER_STRENGTH*t_frame; ship.jets |= 1<<1;}
				if(keystate[SDLK_RIGHT] || keystate[SDLK_n]) { ship.dx += THRUSTER_STRENGTH*t_frame; ship.jets |= 1<<2;}
				if(keystate[SDLK_UP]    || keystate[SDLK_c]) { ship.dy -= THRUSTER_STRENGTH*t_frame; ship.jets |= 1<<3;}
				if(keystate[SDLK_3])		{ SDL_SaveBMP(surf_screen, "snapshot.bmp"); }
			}

			if(keystate[SDLK_p] | keystate[SDLK_s]) {
				if(!pausedown) {
					paused = !paused;
					pausedown = 1;
					if(!paused) ms_end = SDL_GetTicks();
				}
			} else {
				pausedown = 0;
			}
		}

		if(!paused) {
			update_state();

			// SCROLLING
			tmp = (ship.y+ship.h/2 + ship.dy*t_frame - YSCROLLTO)/25 + (ship.dy-screendy);
			screendy += tmp * t_frame/12;
			tmp = (ship.x+ship.w/2 + ship.dx*t_frame - XSCROLLTO)/25 + (ship.dx-screendx);
			screendx += tmp * t_frame/12;
			// taper off so we don't hit the barrier abruptly.
			// (if we would hit in < 2 seconds, adjust to 2 seconds).
			if(dist_ahead + (screendx - BARRIER_SPEED)*TO_TICKS(2) < 0)
				screendx = BARRIER_SPEED - (dist_ahead/TO_TICKS(2));
			dist_ahead += (screendx - BARRIER_SPEED)*t_frame;
			if(MAX_DIST_AHEAD >= 0) dist_ahead = min(dist_ahead, MAX_DIST_AHEAD);

			move_sprites();
			move_dots();

			new_rocks();
			new_engine_dots(&ship);

			collisions();

			// BOUNCE off left or right edge of screen
			if(ship.x < 0 || ship.x+ship.w > XSIZE) {
				ship.x -= (ship.dx-screendx)*t_frame;
				ship.dx = screendx - (ship.dx-screendx)*BOUNCINESS;
			}

			// BOUNCE off top or bottom of screen
			if(ship.y < 0 || ship.y+ship.h > YSIZE) {
				ship.y -= (ship.dy-screendy)*t_frame;
				ship.dy = screendy - (ship.dy-screendy)*BOUNCINESS;
			}

			draw();

			// new game
			if((keystate[SDLK_SPACE] || keystate[SDLK_1] || keystate[SDLK_2])
			   && (state == HIGH_SCORE_DISPLAY
			       || state == TITLE_PAGE
			       || state == GAME_OVER)) {
				if(state == GAME_OVER && new_high_score(score))
					init_score_entry();
				else {
					if((keystate[SDLK_SPACE] && !initial_rocks) || keystate[SDLK_2]) {
						g_easy = 0;
						initial_rocks = NORMAL_I_ROCKS;
						final_rocks = NORMAL_F_ROCKS;
						if(gamespeed == EASY_GAMESPEED)
							gamespeed = NORMAL_GAMESPEED;
					} else if(keystate[SDLK_1]) {
						g_easy = 1;
						initial_rocks = EASY_I_ROCKS;
						final_rocks = EASY_F_ROCKS;
						gamespeed = EASY_GAMESPEED;
					}
					reset_sprites();
					reset_rocks();
					screendx = BARRIER_SPEED; screendy = 0;

					ship.x = XSIZE/2.2; ship.y = YSIZE/2 - ship.w/2;
					ship.dx = screendx; ship.dy = screendy;
					ship.lives = 4;
					ship.flags = MOVE|DRAW|COLLIDE;
					add_sprite(SPRITE(&ship));

					score = 0;

					state = GAMEPLAY;
					play_tune(TUNE_GAMEPLAY);
				}
			}

			ship.jets = 0;
		}

		if(state == TITLE_PAGE && keystate[SDLK_h]) {
			state = HIGH_SCORE_DISPLAY;
			state_timeout = 400;
		}
	}
}

int
main(int argc, char **argv) {
	if(!parse_opts(argc, argv)) return 1;

	if(init()) {
		printf ("vor: SDL error: '%s'\n",initerror);
		return 1;
	}

	start = SDL_GetTicks();
	frames = 0;
	gameloop();
	end = SDL_GetTicks();
	// printf("%ld frames in %ld ms, %.2f fps.\n", frames, end-start, frames * 1000.0 / (end-start));

	return 0;
}
